/* ************************************************************************** */
/*                                       _____           _                    */
/*  libp7/packet.h                      |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/17 18:08:43                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#ifndef LIBP7_PACKET_H
# define LIBP7_PACKET_H
# define MAX_COMMAND_ARGUMENT_SIZE 128

/* ************************************************************************** */
/*  Useful structures                                                         */
/* ************************************************************************** */
/**
 *	p7_version_t:
 *	Version ("xx.xx.xx" + type).
 */

typedef struct {
	int major;
	int minor;
	int rev;
	char type[9];
} p7_version_t;

/**
 *	p7_protocolversion_t:
 *	Protocol version ("xx.x").
 */

typedef struct {
	int major;
	int minor;
} p7_protocolversion_t;

/**
 *	p7_packettype_t:
 *	Packet type.
 */

typedef enum {
	p7_pt_cmd = 0x01,
	p7_pt_data = 0x02,
	p7_pt_roleswp = 0x03,
	p7_pt_check = 0x05,
	p7_pt_ack = 0x06,
	p7_pt_error = 0x15,
	p7_pt_terminate = 0x18
} p7_packettype_t;

/**
 *	p7_packetcommand_t:
 *	Packet command ID.
 */

typedef enum {
	/* Flash commands */
	p7_pt_send = 0x45, /* file transfer */
} p7_commandcode_t;

/**
 *	p7_packetcommandoverwrite_t:
 *	Defines the behaviour if file already exists.
 *
 *	Moo.
 */

typedef enum {
	p7_cow_request_confirmation_before_overwriting = 0x00,
	p7_cow_terminate_if_file_exists = 0x01,
	p7_cow_force_overwrite = 0x02
} p7_commandoverwrite_t;

/**
 *	p7_errcode_t:
 *	Error codes.
 */

typedef enum {
	p7_ec_resend = 0x01,
	p7_ec_overwrite = 0x02,
	p7_ec_overwrite_noresponse = 0x03,
	p7_ec_overwrite_impossible = 0x04,
	p7_ec_fullmem = 0x05
} p7_errcode_t;

/**
 *	p7_termtype_t:
 *	Termination type.
 */

typedef enum {
	p7_tmt_default = 0x00,
	p7_tmt_user = 0x01,
	p7_tmt_timeouts = 0x02,
	p7_tmt_overwrite = 0x03
} p7_termtype_t;

/* ************************************************************************** */
/*  Packet parts specific to one packet type                                  */
/* ************************************************************************** */
/**
 *	p7_packetcommand_t:
 *	Command packet data fields.
 */

typedef struct {
	p7_commandcode_t code;

	p7_commandoverwrite_t overwrite;
	int datatype; /* TODO: find out */
	unsigned int filesize;

	/* i know, a bit big, but i don't want to allocate manually :( */
	char args[6][MAX_COMMAND_ARGUMENT_SIZE];
} p7_packetcommand_t;

/**
 *	p7_packetdata_t:
 *	Data packet data fields.
 */

typedef struct {
	/* total number of data packets, id of the current one */
	unsigned int total, id;

	/* data and data size */
	unsigned char data[256];
	unsigned int size;
} p7_packetdata_t;

/**
 *	p7_packetcheck_t:
 *	Check packet data fields.
 */

typedef struct {
	int initial;
} p7_packetcheck_t;

/**
 *	p7_packetack_t:
 *	Extended ack packet data fields.
 */

typedef struct {
	/* check if is extended */
	int extended;
	/* check if is file overwriting confirmation */
	int overwrite;

	/* preprogrammed ROM info looks wiped out */
	int preprog_rom_wiped;
	/* preprogrammed ROM capacity */
	unsigned int preprog_rom_capacity;
	/* preprogrammed ROM version */
	p7_version_t preprog_rom_version;

	/* flash ROM capacity */
	unsigned int flash_rom_capacity;
	/* RAM capacity */
	unsigned int ram_capacity;

	/* bootcode info looks wiped out */
	int bootcode_wiped;
	/* bootcode version */
	p7_version_t bootcode_version;
	/* bootcode offset */
	unsigned int bootcode_offset;
	/* bootcode_size */
	unsigned int bootcode_size;

	/* OS code version */
	p7_version_t os_version;
	/* OS code offset */
	unsigned int os_offset;
	/* OS code size */
	unsigned int os_size;

	/* protocol version */
	p7_protocolversion_t protocol_version;
	/* product ID */
	char product_id[17];

	/* name set by user in SYSTEM */
	char username[17];
	/* hardware and processor identifiers */
	char hwid[9], cpuid[17];
} p7_packetack_t;

/**
 *	p7_packeterr_t:
 *	Error packet data fields.
 */

typedef struct {
	p7_errcode_t code;
} p7_packeterr_t;

/**
 *	p7_packetterm_t:
 *	Terminate packet data fields.
 */

typedef struct {
	p7_termtype_t reason;
} p7_packetterm_t;

/* ************************************************************************** */
/*  The packet definition                                                     */
/* ************************************************************************** */
/**
 *	p7_packet_t:
 *	Packet representation.
 *
 *	Easy to read.
 */

typedef struct {
	/* main info */
	p7_packettype_t type;

	/* for command packet */
	p7_packetcommand_t command;
	/* for data packet */
	p7_packetdata_t data;
	/* for check packet */
	p7_packetcheck_t check;
	/* for ack packet */
	p7_packetack_t ack;
	/* for error packet */
	p7_packeterr_t error;
	/* for terminate packet */
	p7_packetterm_t term;
} p7_packet_t;

#endif /* LIBP7_PACKET_H */
