/* ************************************************************************** */
/*                                       _____           _                    */
/*  init.c                              |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/16 23:19:46                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <libp7/internals.h>
#define casio_vendor 0x07CF
#define casio_product 0x6101

/* ************************************************************************** */
/*  Global vars                                                               */
/* ************************************************************************** */
/**
 *	p7_error:
 *	The error.
 */

int p7_error = p7_noerror;

/**
 *	p7_active:
 *	Boolean to know if we are in active or passive mode.
 */

int p7_active;

/**
 *	p7_shifted:
 *	Boolean to know if we are shifted.
 */

int p7_shifted;

/**
 *	p7_lusb_context:
 *	libusb context used by libp7.
 */

static libusb_context *p7_lusb_context = NULL;

/**
 *	p7_lusb_handler:
 *	Calculator connexion handler.
 */

libusb_device_handle *p7_lusb_handler = NULL;

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/**
 *	p7_exit:
 *	De-initialize libp7.
 */

static void p7_exit(void)
{
	/* free buffers */
	/* - raw reception packet buffer - */
	if (p7_recv_buffer) free(p7_recv_buffer);
	/* - raw sending packet buffers - */
	if (p7_send_buffers[0]) free(p7_send_buffers[0]);
	if (p7_send_buffers[1]) free(p7_send_buffers[1]);
	/* - decoded packet buffer - */
	if (p7_response) free(p7_response);

	/* close calc handler */
	if (p7_lusb_handler) libusb_close(p7_lusb_handler);

	/* close libusb context */
	if (p7_lusb_context) libusb_exit(p7_lusb_context);
}

/**
 *	p7_init:
 *	Initialize libp7.
 *
 *	Creates the libusb context and connect to the calculator.
 *
 *	@return			if it worked (0 if no)
 */

int p7_init(void)
{
	/* set libp7 exit function */
	atexit(&p7_exit);

	/* open handle */
	log_info("creating libusb context");
	if (libusb_init(&p7_lusb_context)) {
		log_fatal("couldn't create libusb context");
		p7_error = p7_error_libusb;
		return (0);
	}

	/* checking for the calculator */
	log_info("checking for the calculator");
	/* - get device list - */
	libusb_device **device_list;
	int device_count = libusb_get_device_list(p7_lusb_context, &device_list);
	if (device_count < 0) {
		log_error("could not get device list :(");
		p7_error = p7_error_unknown;
		return (0);
	}
	/* - looking for the calc - */
	libusb_device *calc = NULL;
	for (int id = 0; id < device_count; id++) {
		struct libusb_device_descriptor descriptor;
		if (libusb_get_device_descriptor(device_list[id], &descriptor))
			continue;
		if (descriptor.idVendor == casio_vendor
		 && descriptor.idProduct == casio_product) {
			calc = device_list[id];
			break;
		}
	}
	/* - if no calc, snif - */
	if (!calc) {
		log_error("looks like we couldn't find the calculator");
		libusb_free_device_list(device_list, 1);
		p7_error = p7_error_nocalc;
		return (0);
	}

	/* get calculator handler */
	log_info("getting the device handler");
	int err = libusb_open(calc, &p7_lusb_handler);
	libusb_free_device_list(device_list, 1);
	switch (err) {
	/* couldn't get access to device */
	case LIBUSB_ERROR_ACCESS:
		p7_error = p7_error_noaccess;
		return (0);

	/* default case */
	case 0: break;
	default:
		log_fatal("libusb_open returned %d: %s", err,
			libusb_error_name(err));
		log_fatal("errno was %d: %s", errno, strerror(errno));
		p7_error = p7_error_unknown;
		return (0);
	}

	/* allocating buffers */
	if (!(p7_send_buffers[0] = malloc(MAX_PACKET_SIZE)))
		return (0);
	if (!(p7_send_buffers[1] = malloc(MAX_PACKET_SIZE)))
		return (0);
	if (!(p7_recv_buffer = malloc(MAX_PACKET_SIZE)))
		return (0);
	if (!(p7_response = malloc(sizeof(p7_packet_t))))
		return (0);

	/* starting as active */
	p7_active = 1;

	/* initialization went alright */
	return (1);
}
