/* ************************************************************************** */
/*                                       _____           _                    */
/*  packet/send.c                       |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/16 12:43:57                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <libp7/internals.h>

/* ************************************************************************** */
/*  Buffers                                                                   */
/* ************************************************************************** */
/**
 *	p7_send_buffers:
 *	Sending packet buffer.
 */

#define buffer p7_send_buffers[p7_send_buffer_id]
unsigned char *p7_send_buffers[2];

/**
 *	p7_send_buffers_size:
 *	Sending packet buffer sizes.
 */

#define buffer_size p7_send_buffers_size[p7_send_buffer_id]
int p7_send_buffers_size[2];

/**
 *	p7_send_buffer_id:
 *	ID of the current send buffer.
 */

int p7_send_buffer_id = 0;

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/**
 *	p7_send_err_resend:
 *	Send resend request.
 *
 *	In case of bad checksum. As this is a special packet ("meta" packet ?), we
 *	do not want to use up a send buffer for this - packets like this one could
 *	be recursive !
 *
 *	No need for a "special function" like this for receiving, as sending
 *	a "resend" packet means that what we've received was invalid.
 *
 *	@arg	lastbuf			last buffer
 *	@arg	lastbuf_size	last buffer size
 *	@return					if it worked
 */

int p7_send_err_resend(unsigned char *lastbuf, int lastbuf_size)
{
	static unsigned char rpack[] = {p7_pt_error, '0', '1', '0', '6', 'F', 0};

	/* log packet */
	log_info("sending the following resend (basic) packet :");
	logm_info(rpack, 7);

	/* sending loop */
	int wasresend = 0, was_for_last = 1, got_response = 0; int tf;
	unsigned char *buf = rpack; int buf_size = 7;
	do {
		/* log resend */
		if (wasresend) {
			if (p7_shifted) {
				if (was_for_last) {
					buf = lastbuf;
					buf_size = lastbuf_size;
				} else {
					buf = rpack;
					buf_size = 7;
				}
				was_for_last = !was_for_last;
			}
			log_warn("resend request was received, resend it goes");
		}

		/* send prepared packet */
#if LOGLEVEL <= ll_info
		long ts = p7_get_ms();
#endif
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			buf, buf_size, &tf, SENDNG_TIMEOUT);
		log_info("packet sent, sending took %ldµs", p7_get_ms() - ts);

		/* check error */
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}
	} while ((got_response = p7_recv(buf, buf_size))
		&& response.type == p7_pt_error
		&& response.error.code == p7_ec_resend);

	return (got_response);
}

/**
 *	p7_send_basic:
 *	Send a non-extended packet.
 *
 *	@arg	type		the packet type
 *	@arg	subtype		the packet subtype
 *	@arg	resp		check response (is yes for 99.9% of the cases)
 *	@return				if worked (0 if none)
 */

int p7_send_basic(unsigned int type, unsigned int subtype, int resp)
{
	/* check if handler is initialized */
	if (!p7_lusb_handler) {
		p7_error = p7_error_uninitialized;
		return (0);
	}

	/* check if user wants to shift */
	if (!resp) p7_shifted = 1;

	/* change buffer and prepare packet */
	p7_send_buffer_id = !p7_send_buffer_id;
	buffer[0] = (unsigned char)type;
	p7_putascii(&buffer[1], subtype, 2);
	buffer[3] = '0';
	p7_putascii(&buffer[4], p7_checksum(buffer, 6), 2);
	buffer[6] = 0;
	buffer_size = 7;

	/* log packet */
	log_info("sending the following basic packet :");
	logm_info(buffer, buffer_size);

	/* sending loop */
	int wasresend = 0, got_response = 1; int tf;
	do {
		/* resend */
		if (wasresend) {
			if (p7_shifted)
				p7_send_buffer_id = !p7_send_buffer_id;
			log_warn("resend request was received, resend it goes");
		}

		/* send prepared packet */
#if LOGLEVEL <= ll_info
		long ts = p7_get_ms();
#endif
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			buffer, buffer_size, &tf, SENDNG_TIMEOUT);
		log_info("packet sent, sending took %ldµs", p7_get_ms() - ts);

		/* check error */
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}
	} while (resp && (got_response = p7_recv(buffer, buffer_size))
		&& response.type == p7_pt_error
		&& response.error.code == p7_ec_resend);

	/* packet sending is finished */
	return (got_response);
}

/**
 *	p7_send_ext:
 *	Send an extended packet.
 *
 *	@arg	type		the packet type
 *	@arg	subtype		the packet subtype
 *	@arg	data		the packet data
 *	@arg	size		the packet data size
 *	@arg	resp		check response (is yes for 99.9% of the cases)
 *	@return				if it worked
 */

int p7_send_ext(unsigned int type, unsigned int subtype,
	void *data, unsigned int size, int resp)
{
	/* check if handler is initialized */
	if (!p7_lusb_handler) {
		p7_error = p7_error_uninitialized;
		return (0);
	}

	/* check if user wants to shift */
	if (!resp) p7_shifted = 1;

	/* change buffer and prepare packet */
	p7_send_buffer_id = !p7_send_buffer_id;
	/* - first infos - */
	buffer[0] = (unsigned char)type;
	p7_putascii(&buffer[1], subtype, 2);
	buffer[3] = '1';
	/* - data - */
	size = p7_encode(&buffer[8], data, size);
	p7_putascii(&buffer[4], size, 4);
	/* - checksum - */
	p7_putascii(&buffer[8 + size], p7_checksum(buffer, 8 + size + 2), 2);
	buffer[8 + size + 2] = 0;
	buffer_size = 8 + size + 3;

	/* log packet */
	log_info("sending the following extended packet :");
	logm_info(buffer, buffer_size);

	/* sending loop */
	int wasresend = 0, got_response = 1; int transferred;
	do {
		/* log resend request */
		if (wasresend) {
			if (p7_shifted)
				p7_send_buffer_id = !p7_send_buffer_id;
			log_warn("resend request was received, resend it goes");
		}

		/* send prepared packet */
#if LOGLEVEL <= ll_info
		long ts = p7_get_ms();
#endif
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			buffer, buffer_size, &transferred, SENDNG_TIMEOUT);
		log_info("packet sent, sending took %ldµs", p7_get_ms() - ts);

		/* check error */
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}

		/* set wasreset for logging */
		wasresend = 1;
	} while (resp && (got_response = p7_recv(buffer, buffer_size))
		&& response.type == p7_pt_error
		&& response.error.code == p7_ec_resend);

	/* packet sending is finished */
	return (got_response);
}

/**
 *	p7_send_again:
 *	Send the last packet again.
 *
 *	Useful for when resending wasn't managed in one of the previous functions.
 *	Only used for unshifting right now.
 *
 *	@return				if it worked
 */

int p7_send_again(void)
{
	/* check if handler is initialized */
	if (!p7_lusb_handler) {
		p7_error = p7_error_uninitialized;
		return (0);
	}

	/* log packet */
	log_info("sending again the following extended packet :");
	logm_info(buffer, buffer_size);

	/* sending loop */
	int wasresend = 0, got_response = 1; int transferred;
	do {
		/* log resend request */
		if (wasresend) {
			if (p7_shifted)
				p7_send_buffer_id = !p7_send_buffer_id;
			log_warn("resend request was received, resend it goes");
		}

		/* send prepared packet */
#if LOGLEVEL <= ll_info
		long ts = p7_get_ms();
#endif
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			buffer, buffer_size, &transferred, SENDNG_TIMEOUT);
		log_info("packet sent, sending took %ldµs", p7_get_ms() - ts);

		/* check error */
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}

		/* set wasreset for logging */
		wasresend = 1;
	} while ((got_response = p7_recv(buffer, buffer_size))
		&& response.type == p7_pt_error
		&& response.error.code == p7_ec_resend);

	/* packet sending is finished */
	return (got_response);
}
