/* ************************************************************************** */
/*                                       _____           _                    */
/*  send_command.c                      |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/17 09:51:03                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libp7/internals.h>

/*
 *	Types and subtypes
 */

#define T p7_pt_cmd
/* System commands */
/* - restart/reset the equipment - */
#define ST_SYS_RESET 0x00
/* - get device info - */
#define ST_SYS_GETDEVINFO 0x01
/* - set link settings - */
#define ST_SYS_SETLINK 0x02

/* Flash commands */
/* - create directory - */
#define ST_FLS_MKDIR 0x40
/* - delete directory - */
#define ST_FLS_RMDIR 0x41
/* - rename directory - */
#define ST_FLS_MVDIR 0x42
/* - change working directory - */
#define ST_FLS_CWD 0x43
/* - file transfer request - */
#define ST_FLS_REQ 0x44
/* - file transfer - */
#define ST_FLS_SEND 0x45
/* - delete file - */
#define ST_FLS_DEL 0x46
/* - rename file - */
#define ST_FLS_MV 0x47
/* - copy file - */
#define ST_FLS_CP 0x48
/* - file transfer all request - */
#define ST_FLS_RECVALL 0x49
/* - UNKNOWN - reset flash? (DO NOT USE) - */
//#define ST_FLS_RESETFLASH 0x4A
/* - capacity transmit request - */
#define ST_FLS_RECV_CAPACITY 0x4B
/* - capacity transmit - */
#define ST_FLS_SEND_CAPACITY 0x4C
/* - file info transfer all request - */
#define ST_FLS_RECVALL_INFO 0x4D
/* - file info transfer - */
#define ST_FLS_SENDALL_INFO 0x4E
/* - flash image transfer request - */
#define ST_FLS_RECVFLS 0x4F
/* - flash image transfer - */
#define ST_FLS_SENDFLS 0x50
/* - optimize filesystem - */
#define ST_FLS_OPTIMIZE 0x51

/**
 *	p7_last_sent_command:
 *	Code of the last sent command (last command packet subtype).
 */

unsigned int p7_last_sent_command = 0;

/**
 *	p7_send_cmd:
 *	Send command packet without data field.
 *
 *	Used to perform or request actions. The subtype selects the action.
 *
 *	@arg	subtype		the subcommand
 *	@return				if it worked
 */

static int p7_send_cmd(unsigned int subtype)
{
	/* save last sent command */
	p7_last_sent_command = subtype;

	/* send */
	return (p7_send_basic(T, subtype, 1));
}

/**
 *	p7_send_cmd_data:
 *	Send command packet with data field.
 *
 *	Used to perform or request actions. The subtype selects the action.
 *
 *	@arg	subtype		The subcommand
 *	@arg	overwrite	Mode of operation if a file exists
 *						0: request user confirmation before overwriting
 *						1: terminate if file already exists
 *						2: force overwrite
 *	@arg	datatype	MCS data type
 *	@arg	filesize	the filesize
 *	@arg	arg1		first command argument
 *	@arg	arg2		second command argument
 *	@arg	arg3		third command argument
 *	@arg	arg4		fourth command argument
 *	@arg	arg5		fifth command argument
 *	@arg	arg6		sixth command argument
 *	@return				if it worked
 */

static int p7_send_cmd_data(unsigned int subtype, int overwrite, int datatype,
	unsigned int filesize, char *arg1, char *arg2, char *arg3, char *arg4,
	char *arg5, char *arg6)
{
	/* save last sent command */
	p7_last_sent_command = subtype;

	/* get args length */
	size_t l1 = arg1 ? strlen(arg1) : 0;
	size_t l2 = arg2 ? strlen(arg2) : 0;
	size_t l3 = arg3 ? strlen(arg3) : 0;
	size_t l4 = arg4 ? strlen(arg4) : 0;
	size_t l5 = arg5 ? strlen(arg5) : 0;
	size_t l6 = arg6 ? strlen(arg6) : 0;

	/* prepare data buffer */
	unsigned int buflen = (unsigned int)(24 + l1 + l2 + l3 + l4 + l5 + l6);
	unsigned char buf[buflen];

	/* put info */
	p7_putascii(buf, overwrite, 2);
	p7_putascii(&buf[2], datatype, 2);
	p7_putascii(&buf[4], filesize, 8);
	p7_putascii(&buf[12], l1, 2);
	p7_putascii(&buf[14], l2, 2);
	p7_putascii(&buf[16], l3, 2);
	p7_putascii(&buf[18], l4, 2);
	p7_putascii(&buf[20], l5, 2);
	p7_putascii(&buf[22], l6, 2);

	/* put strings */
	unsigned char *b = &buf[24];
	memcpy(b, arg1, l1); b += l1;
	memcpy(b, arg2, l2); b += l2;
	memcpy(b, arg3, l3); b += l3;
	memcpy(b, arg4, l4); b += l4;
	memcpy(b, arg5, l5); b += l5;
	memcpy(b, arg6, l6);

	/* send */
	return (p7_send_ext(T, subtype, buf, buflen, 1));
}

/**
 *	p7_send_cmdsys_reset:
 *	Send reset command packet.
 *
 *	@return				if it worked
 */

int p7_send_cmdsys_reset(void)
{
	return (p7_send_cmd(ST_SYS_RESET));
}

/**
 *	p7_send_cmdsys_getinfo:
 *	Send "get device info" command packet.
 *
 *	@return				if it worked
 */

int p7_send_cmdsys_getinfo(void)
{
	return (p7_send_cmd(ST_SYS_GETDEVINFO));
}

/**
 *	p7_send_cmdsys_setlink:
 *	Send "set link" command packet.
 *
 *	TODO: make a better function or remove it.
 *
 *	@arg	baudrate	the baud rate (e.g. 9600 or 19200)
 *	@arg	parity		the parity :
 *						0: NONE
 *						1: ODD
 *						2: EVEN
 *	@arg	stopbits	Stop bits, 1 or 2.
 *	@return				if it worked
 */

int p7_send_cmdsys_setlink(unsigned int baudrate, int parity, int stopbits)
{
	/* get strings out of all of this */
	char arg1[5] = "0000", arg2[5] = "", arg3[2] = "";
	
	p7_putascii((unsigned char*)arg1, baudrate, 4);
	sprintf(arg2, !parity ? "NONE" : parity % 2 ? "ODD" : "EVEN");
	sprintf(arg3, "%d", stopbits);

	/* then send packet */
	return (p7_send_cmd_data(ST_SYS_SETLINK, 0, 0, 0,
		arg1, arg2, arg3, NULL, NULL, NULL));
}

/**
 *	p7_send_cmdfls_reqfile:
 *	Request file transfer.
 *
 *	@arg	dirname		the name of the directory
 *	@arg	filename	the name of the file
 *	@arg	devname		the name of the device (usually "fls0")
 *	@return				if it worked
 */

int p7_send_cmdfls_reqfile(char *dirname, char *filename, char *devname)
{
	/* send packet */
	return (p7_send_cmd_data(ST_FLS_REQ, 0, 0, 0,
		dirname, filename, NULL, NULL, devname, NULL));
}

/**
 *	p7_send_cmdfls_sendfile:
 *	Transfers a file.
 *
 *	@arg	ow			the overwrite mode
 *	@arg	fs			the filesize
 *	@arg	dirname		the name of the directory
 *	@arg	filename	the name of the file
 *	@arg	devname		the name of the device (usually "fls0")
 *	@return				if it worked
 */

int p7_send_cmdfls_sendfile(p7_commandoverwrite_t ow, unsigned int fs,
	char *dirname, char *filename, char *devname)
{
	/* send packet */
	return (p7_send_cmd_data(ST_FLS_SEND, ow, 0, fs,
		dirname, filename, NULL, NULL, devname, NULL));
}

/**
 *	p7_send_cmdfls_delfile:
 *	Deletes a file.
 *
 *	@arg	dirname		the name of the directory
 *	@arg	filename	the name of the file
 *	@arg	devname		the name of the device (usually "fls0")
 *	@return				if it worked
 */

int p7_send_cmdfls_delfile(char *dirname, char *filename, char *devname)
{
	/* send packet */
	return (p7_send_cmd_data(ST_FLS_DEL, 0, 0, 0,
		dirname, filename, NULL, NULL, devname, NULL));
}

/**
 *	p7_send_cmdfls_copyfile:
 *	Copy a file.
 *
 *	@arg	dirname		the name of the directory
 *	@arg	filename	the name of the file
 *	@arg	newdir		the name of the new directory
 *	@arg	newname		the new name of the file
 *	@arg	devname		the device name
 *	@return				if it worked
 */

int p7_send_cmdfls_copyfile(char *dirname, char *filename, char *newdir,
	char *newname, char *devname)
{
	/* send packet */
	return (p7_send_cmd_data(ST_FLS_CP, 0, 0, 0,
		dirname, filename, newdir, newname, devname, NULL));
}
