/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/reqfile.c                  |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/20 22:14:42                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdlib.h>
#include <string.h>
#include <libp7/internals.h>
/* number of buffers it will write in the file at a time */
#define BUFNUM 1024

/**
 *	p7_reqfile:
 *	Requests a file.
 *
 *	@arg	file		the file (should be writable)
 *	@arg	dirname		the directory name
 *	@arg	filename	the filename
 *	@arg	devname		the device name
 *	@return				if it worked
 */

int p7_reqfile(FILE *file, char *dirname, char *filename, char *devname)
{
	int ret = 0;

	/* check filename */
	if (!p7_validate_filename(filename)) {
		p7_error = p7_error_filename;
		log_error("sent filename was crappy, returning error to user");
		return (0);
	}

	/* check dirname */
	if (dirname && !p7_validate_dirname(dirname)) {
		p7_error = p7_error_dirname;
		log_error("sent dirname was crappy, returning error to user");
		return (0);
	}

	/* dynamically allocate buf size */
	unsigned int bufsize = MAX_RAWDATA_SIZE * BUFNUM;
	unsigned char *buf = malloc(bufsize);
	if (!buf) {
		p7_error = p7_error_internal;
		log_fatal("Couldn't allocate file buffer");
		return (0);
	}

	/* send command packet */
	log_info("sending file transfer request");
	if (!p7_send_cmdfls_reqfile(dirname, filename, devname)) {
		log_fatal("couldn't send file transfer request/didn't receive answer");
		goto end;
	}

	/* check response packet */
	if (response.type == p7_pt_error
	 && response.error.code == p7_ec_overwrite_impossible) {
		p7_error = p7_error_overwrite_impossible;
		log_fatal("wrong device name or requested file doesn't exist");
		goto end;
	} else if (response.type != p7_pt_ack) {
		p7_error = p7_error_unknown;
		log_fatal("didn't receive ack or known error...");
		goto end;
	}

	/* swap roles */
	log_info("sending roleswap");
	if (!p7_send_roleswp()) {
		log_fatal("couldn't swap roles");
		goto end;
	}

	/* receive response file transfer command */
	log_info("checking obtained command");
	if (response.type != p7_pt_cmd) {
		log_fatal("didn't receive expected command");
		goto end;
	}

	/* - Note: we are now in passive mode - */
	/* responding with acks */
	log_info("starting main loop");
	unsigned char *b = buf; unsigned int fillsize = 0;
	p7_send_ack(0);
	while (1) {
		/* send ack */
		log_info("send ack");
		if (!p7_send_ack(1)) {
			log_fatal("couldn't send ack/didn't receive answer");
			goto end;
		}

		/* if packet is not data or roleswap, wtf casio ? */
		if (response.type != p7_pt_data && response.type != p7_pt_roleswp) {
			p7_error = p7_error_unknown;
			log_fatal("packet type was unplanned, wtf ?");
			goto end;
		}

		/* if packet is roleswap, write what's left and break loop */
		if (response.type == p7_pt_roleswp) {
			log_info("is roleswap, emptying buffer in file");
			fwrite(buf, 1, fillsize, file);
			log_info("our role as passive is now done");
			break;
		}

		/* otherwise (data), continue */
		log_info("is data, put it in buffer");
		memcpy(b, response.data.data, response.data.size);
		fillsize += response.data.size, b += response.data.size;
		if (fillsize > bufsize - MAX_RAWDATA_SIZE) {
			log_info("buffer too full, should be emptied");
			/* empty buffer in file */
			fwrite(buf, 1, fillsize, file);
			b = buf, fillsize = 0;
		}
	}
	/* unshift */
	p7_unshift();

	ret = 1;
end:
	/* free buffer and return */
	free(buf);
	return (ret);
}
