/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/sendfile.c                 |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/19 14:55:33                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdlib.h>
#include <string.h>
#include <libp7/internals.h>
/* number of buffers it will read from the file at a time */
#define BUFNUM 1024

/**
 *	p7_sendstream:
 *	Sends a stream.
 *
 *	Some I/O subtility at the end, as we shouldn't count on local I/O rapidity.
 *
 *	@arg	file		the file stream
 *	@arg	size		the file size
 *	@arg	dirname		the directory name
 *	@arg	filename	the file name
 *	@arg	devname		the device name
 *	@arg	overwrite	if no confirmation callback, should overwrite or not
 *	@arg	ow_conf		if file exists and this callback is not NULL,
 *						call it. If answer is 1, overwrite, otherwise, not.
 *	@return				if it worked
 */

int p7_sendstream(FILE *file, unsigned int size, char *dirname, char *filename,
	char *devname, int overwrite, int (*ow_conf)(void))
{
	int ret = 0;

	/* check filename */
	if (!p7_validate_filename(filename)) {
		p7_error = p7_error_filename;
		log_error("sent filename was crappy, returning error to user");
		return (0);
	}

	/* check dirname */
	if (dirname && !p7_validate_dirname(dirname)) {
		p7_error = p7_error_dirname;
		log_error("sent dirname was crappy, returning error to user");
		return (0);
	}

	/* check if fits in 2 Mo -- so that it sticks for sure in 32 bits */
	if (size >= 2097152) {
		p7_error = p7_error_fullmem;
		log_error("sent file is too large, returning error to user");
		return (0);
	}

	/* get beginning */
#if LOGLEVEL <= ll_info
	long begin = p7_get_ms();
#endif

	/* dynamically allocate buf size */
	unsigned int bufsize = MAX_RAWDATA_SIZE * BUFNUM;
	bufsize = size < bufsize ? size : bufsize;
	unsigned char *buf = malloc(8 + bufsize);
	if (!buf) {
		p7_error = p7_error_internal;
		log_fatal("Couldn't allocate file buffer");
		return (0);
	}

	/* choose overwrite mode
	 *	If we want to never overwrite, we don't want the calculator to terminate
	 *	communication, so we ask confirmation and we will systematically
	 *	return no.
	 */
	p7_commandoverwrite_t owmode;
	if (ow_conf || !overwrite)
		owmode = p7_cow_request_confirmation_before_overwriting;
	else owmode = p7_cow_force_overwrite;

	/* send command packet */
	log_info("sending file transfer command");
	if (!p7_send_cmdfls_sendfile(owmode, size, dirname, filename, devname)) {
		log_fatal("couldn't send file transfer packet/didn't receive answer");
		goto end;
	}

	/* get response packet */
	log_info("checking file transfer answer");
	if (response.type != p7_pt_ack && response.type != p7_pt_error) {
		p7_error = p7_error_unknown;
		log_fatal("file transfer answer was unknown");
		goto end;
	}

	/* - checking what happened if it is an error packet - */
	if (response.type == p7_pt_error) {
		switch (response.error.code) {
		/* file exists, confirm or not */
		case p7_ec_overwrite:
			log_info("calc wanted overwrite confirmation");
			/* check whether to ask */
			int denied = 0;
			if (ow_conf && (*ow_conf)()) {
				log_info("sending yes to overwrite confirmation");
				if (!p7_send_ack_confirm_ow())
					goto end;
			} else {
				log_info("sending no to overwrite confirmation");
				if (!p7_send_err_overwrite_no())
					goto end;
				denied = 1;
				p7_error = p7_error_denied_overwrite;
			}
			/* check if response packet is ack */
			if (response.type != p7_pt_ack) {
				p7_error = p7_error_unknown;
				log_info("response packet wasn't ack");
				goto end;
			}
			/* if we denied, stop here
			 *	not really an error, but this is to say to the user that
			 *	file transfer did not happen.
			 *
			 *	we used the "denied" var instead of p7_error directly as
			 *	p7_error might be uninitialised and contain
			 *	p7_error_denied_overwrite from a previous call to sendfile.
			 */
			if (denied) {
				p7_error = p7_error_denied_overwrite;
				goto end;
			}
			break;

		/* impossible overwrite */
		case p7_ec_overwrite_noresponse:
		case p7_ec_overwrite_impossible:
			p7_error = p7_error_overwrite_impossible;
			log_error("overwrite impossible");
			goto end;

		/* memory full */
		case p7_ec_fullmem:
			p7_error = p7_error_fullmem;
			log_error("distant memory is full");
			goto end;

		/* resend error - just here to stfu the warning */
		case p7_ec_resend:
			p7_error = p7_error_unknown;
			goto end;
		}
	}

	/* get total and last packet size */
	unsigned int lastpacket_size = size % MAX_RAWDATA_SIZE;
	unsigned int total = size / MAX_RAWDATA_SIZE
		+ !!lastpacket_size; /* left data */
	if (!lastpacket_size) lastpacket_size = MAX_RAWDATA_SIZE;

	/* calculate file info */
	log_info("about to send %u packets", total);
	if (lastpacket_size) {
		log_info("last one is irregular and contains %uo", lastpacket_size);
	}

	/* send data loop */
	unsigned int datasize = MAX_RAWDATA_SIZE;
	for (unsigned int id = 1; id <= total;) {
		/* read big block */
		fread(8 + buf, 1, bufsize, file);

		/* then send each block */
		unsigned char *b = buf;
		unsigned int bufnum = BUFNUM;
		for (; bufnum-- && id <= total; id++) {
			/* get datasize */
			if (id == total) datasize = lastpacket_size;

			/* logging */
			log_info("sending packet %u/%u (%uo)", id, total, datasize);

			/* send data packet */
			if (!p7_send_data_quicker(total, id, b, datasize, (id == 1))) {
				log_error("could not send data/receive response");
				goto end;
			}

			/* - check response - */
			if (response.type != p7_pt_ack) {
				p7_error = p7_error_unknown;
				log_error("calculator didn't send ack, weird");
				goto end;
			}

			/* increment buf pointer */
			b += MAX_RAWDATA_SIZE;
		}
	}

	/* unshift */
	p7_unshift();

	/* put finishing time */
	log_info("Sending completed in %ldµs", p7_get_ms() - begin);

	ret = 1;
end:
	/* free buffer and return */
	free(buf);
	return (ret);
}

/**
 *	p7_sendfile:
 *	Sends a file.
 *
 *	@arg	file		the file stream (must be seekable !)
 *	@arg	dirname		the directory name
 *	@arg	filename	the file name
 *	@arg	devname		the device name
 *	@arg	overwrite	if no confirmation required, should overwrite or not
 *	@arg	ow_conf		if file exists and this callback is not NULL,
 *						call it. If answer is 1, overwrite, otherwise, not.
 *	@return				if it worked
 */

int p7_sendfile(FILE *file, char *dirname, char *filename,
	char *devname, int overwrite, int (*ow_conf)(void))
{
	/* calculate file size */
	log_info("calculating given file size");
	unsigned int size;
	if (fseek(file, 0, SEEK_END)) {
		p7_error = p7_error_noseek;
		log_error("could not seek in file");
		return (0);
	}
	size = (unsigned int)ftell(file);
	rewind(file);
	/* - log - */
	log_info("filesize is %uo", size);

	/* send file */
	return (p7_sendstream(file, size, dirname, filename, devname,
		overwrite, ow_conf));
}
